"""
This layer of submodules provides the application interfaces for various use cases.
Some of them are intended to provide a compatibility layer for the existing libraries for other languages and programming systems.
"""
